//! This module contains the core services (and channel definitions) that make up a `fleck` node.
//!
//! Currently, the following externally-accessible services are added by default:

pub mod crypto;
pub mod discovery;
pub mod io;
pub mod msg;
pub mod node;
pub mod peer;
pub mod routing;

pub use msg::{Message, MessageChannel, MessageParams, MessageService};
pub use node::{Node, NodeRegistrationChannel, NodeService};
pub use peer::{Peer, PeerService};

pub mod channel_tags {
    pub struct SendPacketTag {}
    pub struct ReceivePacketTag {}
    pub struct MinorTickTag {}
    pub struct MajorTickTag {}
}

#[derive(PartialEq, PartialOrd, Clone)]
pub enum SendOrder {
    Route,
    Sign,
    Encrypt,
    Dispatch,
}
pub type SendPacketChannel = (channel_tags::SendPacketTag, msg::Message, SendOrder);

#[derive(PartialEq, PartialOrd, Clone)]
pub enum ReceiveOrder {
    /// Annotate with various properties based on message source
    Annotate,
    Decrypt,
    Verify,
    Parse,
    Dispatch,
}
pub type ReceivePacketChannel = (channel_tags::ReceivePacketTag, msg::Message, ReceiveOrder);

/// Minor ticks (roughly once per second): for retries, status updates, etc.
pub type MinorTickChannel = (channel_tags::MinorTickTag, ());
/// Major ticks (several times per minute): for retries, cleanup tasks, periodic queries, etc
pub type MajorTickChannel = (channel_tags::MajorTickTag, ());

/// Initial bootstrapping service to create the other core services.
#[derive(Default)]
pub struct CoreInitService {}

impl crate::service::DefaultService for CoreInitService {
    fn setup(self: &std::rc::Rc<Self>, api: std::rc::Rc<crate::API>) {
        api.create_channel::<MinorTickChannel>();
        api.create_channel::<MajorTickChannel>();
        let io_api = api.clone();
        api.with_service(move |io: &io::IOService| {
            let minor_api = io_api.clone();
            let major_api = io_api;
            io::TimerFdBuilder::default()
                .interval(std::time::Duration::from_secs(1))
                .trigger(move || minor_api.queue::<MinorTickChannel>(()))
                .build(io);
            io::TimerFdBuilder::default()
                .interval(std::time::Duration::from_secs(15))
                .trigger(move || major_api.queue::<MajorTickChannel>(()))
                .build(io);
        });

        api.add_service::<msg::MessageService>();
        api.add_service::<node::NodeService>();
        api.add_service::<discovery::DiscoveryService>();
        api.add_service::<discovery::LocalDiscovery>();
        api.add_service::<discovery::PeerDiscovery>();
        api.add_service::<peer::PeerService>();
        api.add_service::<crypto::SignPacket>();
        api.add_service::<crypto::EncryptionService>();
        api.add_service::<routing::RoutingTableService>();
        api.add_service::<routing::PacketRoutingService>();
    }
}
